from .poly_base import BasePolynomial
from .poly_list import ListPolynomial
import numpy as np
from fractions import Fraction

class PointsPolynomial(BasePolynomial):
    """ Class which creates object of polynomial 
        in the form of points. Implemented addition, 
        subtraction and multiplication in this form, 
        the rest is based on interpolation to the form of a list of coefficients.
    """
    def __init__(self, poly):
        """ Creating object of this class. Parsing string to form of ListPolynomial,
            using Horner's Method to calculating points.
        """
        super().__init__()
        if isinstance(poly, str):
            listPolynomial = ListPolynomial(poly)
            self._polynomial = [(x, listPolynomial(x)) for x in range(len(listPolynomial._polynomial))]
        elif isinstance(poly, list):
            self._polynomial = poly
        else:
            raise TypeError("Polynomials must be str or list of points")

    # def toList(self): # zamiana z punktów na listę wspolczynnikow
    #     degree = len(self._polynomial)
    #     x = np.array(self._polynomial)
    #     eq = [[i**p for p in range(degree)] for i in range(degree)] # rownania
    #     Y = x[:, 1] # prawa strona ukladu rownan
    #     X = np.array(eq)
    #     A = np.linalg.solve(X, Y) # wyznaczenie wspolczynnikow
    #     return [int(np.round(i)) for i in list(A)]

    def __extendForm(self, other):
        """ Increases the number of polynomial points so that you can make calculations on them, 
            the number of points must be the same for the calculation.
        """
        length1 = len(self._polynomial)
        length2 = len(other._polynomial)
        if length1 < length2:
            poly = self.__extend(length1, length2)
            return poly, other
        else:
            poly = other.__extend(length2, length1)
            return self, poly

    def __extend2n(self):
        """ Increases the form to 2n points for multiplication
        """
        length = len(self._polynomial)
        poly = self.__extend(length, 2*length)
        return poly

    def __extend(self, length1, length2):
        """ Supplementing list with additional points
        """
        x = self.lagrangeInterpolation()
        newPoly = self._polynomial.copy()
        for i in range(length1, length2):
            newPoly.append((i, x(i)))
        return PointsPolynomial(newPoly)

    def __str__(self):
        """ Creating string from polynomial
            when calling str() or print().
        """
        return str(self.lagrangeInterpolation())

    def __add__(self, other):
        """ Addition of two polynomials and creation of 
            new object of this class.
        """
        if len(self._polynomial) != len(other._polynomial):
            poly1, poly2 = self.__extendForm(other)
        else:
            poly1 = PointsPolynomial(self._polynomial.copy())
            poly2 = PointsPolynomial(other._polynomial.copy())
        return PointsPolynomial([(x[0], x[1] + y[1]) for x, y in zip(poly1._polynomial, poly2._polynomial)])

    def __sub__(self, other):
        """ Subtraction of two polynomials and creation of 
            new object of this class.
        """
        if len(self._polynomial) != len(other._polynomial):
            poly1, poly2 = self.__extendForm(other)
        else:
            poly1 = PointsPolynomial(self._polynomial.copy())
            poly2 = PointsPolynomial(other._polynomial.copy())
        return PointsPolynomial([(x[0], x[1] - y[1]) for x, y in zip(poly1._polynomial, poly2._polynomial)])

    def __mul__(self, other):
        """ Multiplication of two polynomials and creation of 
            new object of this class.
        """
        if len(self._polynomial) != len(other._polynomial):
            poly1, poly2 = self.__extendForm(other)
        else:
            poly1 = PointsPolynomial(self._polynomial.copy())
            poly2 = PointsPolynomial(other._polynomial.copy())
        poly1 = poly1.__extend2n()
        poly2 = poly2.__extend2n()
        return PointsPolynomial([(x[0], x[1] * y[1]) for x, y in zip(poly1._polynomial, poly2._polynomial)])

    def __divmod__(self, other):
        """ Division with the rest of two polynomials and creation of 
            new object of this class.
        """
        poly1 = self.lagrangeInterpolation()
        poly2  = other.lagrangeInterpolation()
        result, r = divmod(poly1, poly2)
        return PointsPolynomial(str(result)), PointsPolynomial(str(r))

    def __call__(self, number):
        """ Method overrides calling class. Calculates value of polynomial
            in specific point using Horner's Method.
        """
        return self.lagrangeInterpolation()(number)

    def derivative(self):
        """ Calculating derivative of polynomial.
        """
        return PointsPolynomial(str(self.lagrangeInterpolation().derivative()))

    def roots(self):
        """ Main method for calculating roots of polynomial.
            Returns list of rational roots.
        """
        return self.lagrangeInterpolation().roots()

    def factorization(self):
        """ Main method for factorization of polynomials. Returns list
            of objects of this class, which are the factors of polynomial.
        """
        factors = self.lagrangeInterpolation().factorization()
        return [PointsPolynomial(str(factor)) for factor in factors]

    def lagrangeInterpolation(self):
        """ Lagrange interpolation for changing form from list of points
            to list of coefficients.
        """
        polyInterpolated = ListPolynomial([0])
        polyMulti = ListPolynomial([1])
        for xy in self._polynomial:
            polyMulti = ListPolynomial([xy[1]])
            for xy2 in self._polynomial:
                if xy[0] == xy2[0]:
                    continue
                currentPoly = ListPolynomial([-xy2[0], 1])
                polyMulti = polyMulti * currentPoly
                polyMulti._polynomial[:] = [x * Fraction(1, xy[0] - xy2[0]) for x in polyMulti._polynomial]
            polyInterpolated = polyInterpolated + polyMulti
        newPoly = [int(np.round(x.numerator/x.denominator)) for x in polyInterpolated._polynomial]
        polyInterpolated = ListPolynomial(newPoly)
        return polyInterpolated


import re
from .sieve_init import *
import math as mt

class BasePolynomial:
    """ Class containing common methods for 3 forms of polynomials.
    """
    def __getFactorization(self, x):
        """ Returns factorization for number and amount of appearances for each factor.
        """
        ret = list()
        while x != 1:
            ret.append(spf[x])
            x = x // spf[x]
        ret += self.primeFactors
        countFactors = [[i,ret.count(i)] for i in set(ret)]
        return countFactors

    def __generateDivisors(self, curIndex, curDivisor, arr):
        """ Recursive generation of divisors.
        """
        if (curIndex == len(arr)):
            self.divisors.append(curDivisor)
            return
        
        for i in range(arr[curIndex][1] + 1):
            self.__generateDivisors(curIndex + 1, curDivisor, arr)
            curDivisor *= arr[curIndex][0]

    def __divisorsForSmallerNumbers(self, n):
        """ Divisors for numbers smaller than MAX_N
        """
        factors = self.__getFactorization(n)
        self.__generateDivisors(0, 1, factors)

    def __divisorsForBiggerNumbers(self, n):
        """ Divisors for numbers bigger than MAX_N
        """
        ifContinue = True
        while ifContinue:
            gen = (i for i in primeNumbers if i <= int(mt.sqrt(n)))
            for p in gen:
                if n%p == 0:
                    self.primeFactors.append(p) # dodajemy najmniejszy dzielnik pierwszy dla dużej liczby 
                    n = n // p
                    ifContinue = True
                    if n <= MAX_N: # jeśli liczba jest mniejsza od MAX_N to działamy na wersji funkcji dla małych liczb
                        self.__divisorsForSmallerNumbers(n)
                        return
                    break
                else:
                    ifContinue = False
        return n
        
    def findDivisors(self, n):
        """ Main function for finding divisors of numbers
        """
        if n != 0:
            n = abs(n)
            self.divisors = []
            self.primeFactors = []
            if n <= MAX_N:
                self.__divisorsForSmallerNumbers(n)
            else:
                newN = self.__divisorsForBiggerNumbers(n)

                if self.divisors == []:
                    self.primeFactors.append(newN)
                    self.primeFactors = [[i,self.primeFactors.count(i)] for i in set(self.primeFactors)]
                    self.__generateDivisors(0, 1, self.primeFactors)
            minus_divisors = [-i for i in self.divisors]
            self.divisors += minus_divisors
            return sorted(self.divisors)
        else:
            return []

    def parsePolynomial(self, poly):
            """ Method which parse polynomial in form of str to one of three
                implemented forms
            """
            pattern = r"\-?[0-9]*x?\^?[0-9]*"
            found = re.findall(pattern, poly)
            patternX = re.compile(r"\-?[0-9]*(?=x)") # znajdywanie liczby przed x
            patternAfterX = re.compile(r"(?<=x\^)[0-9]*") # znajdywanie liczby po x
            coefsValues = {}
            for f in found:
                x = patternX.match(f)
                x2 = patternAfterX.search(f)
                if x and x2: # mamy do czynienia z ax^b
                    if int(x2.group(0)) in coefsValues.keys():
                        raise ValueError("Every monomial must have different power")
                    if x.group(0) == '':
                        coefsValues[int(x2.group(0))] = 1
                    elif x.group(0) == '-':
                        coefsValues[int(x2.group(0))] = -1
                    else:
                        coefsValues[int(x2.group(0))] = int(x.group(0))
                elif x: # mamy do czynienia z ax
                    if 1 in coefsValues.keys():
                        raise ValueError("Every monomial must have different power")
                    if x.group(0) == '':
                        coefsValues[1] = 1
                    elif x.group(0) == '-':
                        coefsValues[1] = -1
                    else:
                        coefsValues[1] = int(x.group(0))
                elif f != '': # mamy do czynienia z liczbą
                    if 0 in coefsValues.keys():
                        raise ValueError("Every monomial must have different power")
                    coefsValues[0] = int(f) 
            return coefsValues
    
    def backParse(self, p, result, length):
        """ Parsing from polynomial form to str
        """
        elem = self._polynomial[p]
        act = ""
        if elem != 0:
            if p != length:
                if elem < 0:
                    if elem != -1 or p == 0:
                        act += str(elem)
                    else:
                        act += "-"
                elif elem != 1 or p == 0:
                    act += "+" + str(elem)
                else:
                    act += "+"
            elif elem != 1:
                if elem == -1:
                    act += "-"
                else:
                    act += str(elem)
            if p > 1:
                act += "x^" + str(p)
            elif p == 1:
                act += "x"            
            
        result += act
        return result

    def __floordiv__(self, other):
        """ Division of polynomials without rest
        """
        poly, _ = divmod(self, other)
        return poly

    def __mod__(self, other):
        """ Modulo operation which returns the rest
            of the polynomial division
        """
        _, r = divmod(self, other)
        return r


import math as mt

MAX_N = 2**16 + 1
spf = [0 for i in range(MAX_N)]
primeNumbers = []

def sieve(MAX_N):
    """ Completes the list with prime numbers up to MAX_N 
        and determines the smallest prime factor of a number
    """
    spf[1] = 1
    for i in range(2, MAX_N):	
        spf[i] = i

    for i in range(4, MAX_N, 2):
        spf[i] = 2

    for i in range(3, mt.ceil(mt.sqrt(MAX_N))):
        if (spf[i] == i):
            for j in range(i * i, MAX_N, i):
                if (spf[j] == j):
                    spf[j] = i

    for i in range(2, MAX_N):
        if i == spf[i]:
            primeNumbers.append(i)

sieve(MAX_N)


from fractions import Fraction
import numpy as np
import math as mt
import itertools
from .poly_base import BasePolynomial

class ListPolynomial(BasePolynomial):
    """ Class which creates object of polynomial 
        in the form of list of coefficients.
    """
    def __init__(self, poly):
        """ Parsing polynomial and creating object of this form.
            Deleting redundant zeros.
        """
        super().__init__()
        if isinstance(poly, str):
            coefsValues = self.parsePolynomial(poly)
            listCoefficients = [0 if i not in coefsValues.keys() else coefsValues[i] for i in range(max(coefsValues.keys())+1)]
            self._polynomial = listCoefficients
        elif isinstance(poly, list):
            while poly and poly[-1] == 0:
                poly.pop()
            self._polynomial = poly
        else:
            raise TypeError('Polynomials must be str or list')

    def __add__(self, other):
        """ Addition of two polynomials and creation of 
            new object of this class.
        """
        return ListPolynomial([x + y for x, y in itertools.zip_longest(self._polynomial, other._polynomial, fillvalue=0)])

    def __sub__(self, other):
        """ Subtraction of two polynomials and creation of 
            new object of this class.
        """
        return ListPolynomial([x - y for x, y in itertools.zip_longest(self._polynomial, other._polynomial, fillvalue=0)])

    def __mul__(self, other):
        """ Multiplication of two polynomials and creation of 
            new object of this class.
        """
        degree1 = len(self._polynomial) - 1 
        degree2 = len(other._polynomial) - 1
        result = [0]*(degree1+degree2+1) # utworzenie listy o rzędzie wynikającym z rzędów pozostałych wielomianów
        for i, p1 in enumerate(self._polynomial):
            for j, p2 in enumerate(other._polynomial):
                result[i+j] += p1*p2 # odpowiednio zmieniamy współczynniki (potęgi dodajemy za potęgi robią tu indexy)
        return ListPolynomial(result)

    def __divmod__(self, other):
        """ Division with the rest of two polynomials and creation of 
            new object of this class.
        """
        poly1length = len(self._polynomial)
        poly2length = len(other._polynomial)
        if poly1length >= poly2length: # jeśli rząd pierwszego wielomianu jest większy bądź równy drugiemu to tworzona jest lista wynikowa
            if other._polynomial == [0]:
                raise ZeroDivisionError("Polynomial Division by zero")
            result = ListPolynomial([])
            # result = [0]*(poly1length-poly2length+1)
        else:
            return ListPolynomial("0"), self

        r = self # inicjalizacja reszty
        i = poly1length 
        polyR = r._polynomial
        while i-poly2length >= 0: 
            p1 = polyR[-1] 
            if p1 % other._polynomial[-1] == 0: # sprawdzamy czy elementy przy najwyższych potęgach są podzielne
                new = [0] * (i-poly2length + 1)
                new[i-poly2length] = int(p1/other._polynomial[-1]) # odpowiedniemu indeksowi w wyniku przypisujemy dzielenie dwóch elementów o najwyższym rzędzie
                result = result + ListPolynomial(new)
                bottom = ListPolynomial(new) * other # mnożymy aktualny rezultat przez dzielnik
                r = r - bottom # odejmujemy od reszty powyższy wymnożony wielomian
                polyR = r._polynomial
                while polyR!=[] and polyR[-1]==0:
                    polyR.pop(-1)
                i = len(polyR)
            else:
                raise ValueError("Polynomials are indivisible")
        return result, r

    def __str__(self):
        """ Creating string from polynomial
            when calling str() or print().
        """
        result = ""
        length = len(self._polynomial)
        if length == 1:
            return str(self._polynomial[0])
        
        # ustawiane są odpowiednie współczynniki wielomianu i zapisywane wraz z x do str
        for p in range(len(self._polynomial)-1, -1, -1):
            result = self.backParse(p, result, length - 1)
        if result == "":
            return "0"

        return result

    def __call__(self, number):
        """ Method overrides calling class which calculate value of polynomial
            in specific point using Horner's Method.
        """
        l = len(self._polynomial) - 1
        result = self._polynomial[l]
        for i in range(l-1 , -1, -1):
            result = number*result + self._polynomial[i] # schemat Hornera
        return result

    def derivative(self):
        """ Derivative of polynomial.
        """
        if len(self._polynomial) == 1:
            return ListPolynomial("0")
        else:
            return ListPolynomial([self._polynomial[i]*i for i in range(1, len(self._polynomial))])

    def __sub_roots(self, result, p, q, quotient_prev):
        """ Auxiliary function for calculating roots.
        """
        for comb in itertools.product(p, q):
            if comb[0]/comb[1] in result or Fraction(comb[0], comb[1]) in result:
                continue
            multi = ListPolynomial([-comb[0], comb[1]])
            try:
                quotient, r = divmod(quotient_prev, multi)
            except:
                continue
            while not r._polynomial: # jeśli reszta z dzielenia ma same zera to znaczy, że badamy krotność pierwiastka
                try:
                    quotient, r = divmod(quotient, multi)
                except:
                    if comb[0] % comb[1] == 0:
                        result.append(int(comb[0]/comb[1]))
                    else:
                        result.append(Fraction(comb[0], comb[1]))
                    break
                if comb[0] % comb[1] == 0:
                    result.append(int(comb[0]/comb[1]))
                else:
                    result.append(Fraction(comb[0], comb[1]))     
        return result 

    def roots(self):
        """ Main method for calculating roots of polynomial.
            Returns list of rational roots.
        """
        result = []
        p = self.findDivisors(self._polynomial[0])
        q = self.findDivisors(self._polynomial[-1])

        if p:
            result = self.__sub_roots(result, p, q, self)                
        else: # w przypadku, gdy wyraz wolny jest 0
            multi = ListPolynomial("x")
            quotient, r = divmod(self, multi) # dzielenie przez (x-a)^n
            while not r._polynomial: # jeśli reszta z dzielenia jest pusta to znaczy, że badamy krotność pierwiastka
                quotient_prev = quotient
                quotient, r = divmod(quotient, multi) # dzielenie przez (x-a)^n 
                result.append(0)
            p = self.findDivisors(quotient_prev._polynomial[0])
            q = self.findDivisors(quotient_prev._polynomial[-1])      
            result = self.__sub_roots(result, p, q, quotient_prev)
        return result

    def factorization(self):
        """ Main method for factorization of polynomials. Returns list
            of objects of this class, which are the factors of polynomial.
        """
        roots = self.roots() # wyznaczenie pierwiastków 
        l = len(roots)
        if l > 0:
            roots = [ListPolynomial([-i, 1]) if not isinstance(i, Fraction) 
            else ListPolynomial([-i.numerator, i.denominator]) for i in roots] # utworzenie wielomianów z pierwiastków

            multi = ListPolynomial([1])
            if l > 1:
                for i in range(l):
                    multi = multi * roots[i] # wymnażanie kolejnych pierwiastków
                factor, _ = divmod(self, multi) # otrzymujemy wielomian poprzez dzielenie
            else:
                factor, _ = divmod(self, roots[0])
            if len(factor._polynomial) < 5:
                if factor._polynomial != [1]: # nie dodajemy 1 jako wyniku dzielenia
                    roots.append(factor) # do pierwiastków dodajemy nierozkłdalną część
            else:
                factors = factor.__no_roots_factorization()
                roots += factors
        else:
            if len(self._polynomial) < 5:
                roots = self # do pierwiastków dodajemy nierozkłdalną część
            else:
                factors = self.__no_roots_factorization()
                roots += factors
        return roots

    def __check_degree(self, list_of_polynomials):
        """ Checks whether the factorization still makes sense, 
            i.e. that the degree must be greater than 3
        """
        for poly in list_of_polynomials:
            if len(poly._polynomial) > 4:
                return True
        return False

    def __no_roots_factorization(self):
        """ Auxiliary function for calculating factorization.
            Implements Kronecker's Method.
        """
        factor_polynomials = [self]
        checked_polynomials = []
        while self.__check_degree(factor_polynomials):
            found = False
            for poly in factor_polynomials: # znajdz wielomian o rzędzie większym niż 3
                if len(poly._polynomial) > 4:
                    currentPoly = poly
                    break
            l = len(currentPoly._polynomial)
            k = mt.floor(l/2) if l % 2 == 0 else mt.floor(l/2) + 1 
            values = [] # kolejne wartosci wielomianu
            equations = [] # rownania w zaleznosci od wybranych punktow
            for i in range(k): # obliczanie macierzy do rownan oraz wartosci wielomianu
                values.append(currentPoly(i))
                eq = [i**p for p in range(k)]
                equations.append(eq)
            A = np.array(equations)
            inv_A = np.linalg.inv(A)
            factors_of_values = []
            for number in values: # wyznaczanie dzielnikow kolejnych wartosci wielomianu
                factors_of_values.append(self.findDivisors(number))
            without_minus = []
            for y in factors_of_values[0]: # usuwanie tylko tych kombinacji ktore roznia sie tylko znakiem
                if y > 0:
                    without_minus.append(y)
            
            factors_of_values[0] = without_minus
            for combination in itertools.product(*factors_of_values):
                combination = list(combination)
                B = np.array(combination) # macierz B potrzebna do znalezienia współczynników wielomianu
                X = inv_A.dot(B) 
                X = [int(np.round(X[i])) for i in range(X.size)]
                while X[-1] == 0:
                    X.pop()

                X = ListPolynomial(X)
                try:
                    factor, r = divmod(currentPoly, X) # sprawdzenie czy wielomian jest dzielnikiem
                except:
                    continue
                if r._polynomial==[] and len(X._polynomial) > 1: # jesli znalezlismy dzielnik to dodajemy go do rozkladu wielomianu wraz z dzielnikiem, który idzie z nim w parze
                    factor_polynomials.append(X)
                    factor_polynomials.append(factor)
                    found = True
                    break
            if found: 
                factor_polynomials.remove(currentPoly) # jeśli znaleźliśmy rozkład do usuwamy wielomian, którym się zajmowaliśmy
            else:
                factor_polynomials.remove(currentPoly) # jeśli brak rozkładu dla wielomianu to usuwamy go z wielomianów do sprawdzenia i dodajemy do już sprawdzonych
                checked_polynomials.append(currentPoly)
        factor_polynomials += checked_polynomials
        return factor_polynomials


from fractions import Fraction
import itertools
from .poly_base import BasePolynomial
import numpy as np
import math as mt
from collections import defaultdict

class DictPolynomial(BasePolynomial):
    """ Class which creates object of polynomial 
        in the form of dict of coefficients.
        For a key to be present, the coefficients must be different from zero.
    """
    def __init__(self, poly):
        """ Parsing polynomial and creating object of this form.
            Deleting redundant zeros.
        """
        super().__init__()
        if isinstance(poly, str):
            coefsValues = self.parsePolynomial(poly)
            self._polynomial = coefsValues
        elif isinstance(poly, dict):
            self._polynomial = {}
            for key in poly.keys():
                if poly[key] != 0:
                    self._polynomial[key] = poly[key]
        else:
            raise TypeError('Polynomials must be str or dict')    

    def __str__(self):
        """ Creating string from polynomial
            when calling str() or print().
        """
        result = ""
        if self._polynomial != {}:
            length = max(self._polynomial.keys())
        else:
            return "0"

        if length == 0:
            return str(self._polynomial[0])

        for p in sorted(self._polynomial.keys(), reverse=True):
            result = self.backParse(p, result, length)
        if result == "":
            return "0"

        return result

    def __add__(self, other):
        """ Addition of two polynomials and creation of 
            new object of this class.
        """
        keys_result = list(self._polynomial.keys()) + list(other._polynomial.keys()) # weź wszystkie klucze dla dwóch wielomianów
        result = dict.fromkeys(keys_result) # usuń duplikaty kluczy i stwórz słownik dla wyniku
        for key in result.keys():
            result[key] = self._polynomial.get(key, 0) + other._polynomial.get(key, 0)
        return DictPolynomial(result)

    def __sub__(self, other):
        """ Subtraction of two polynomials and creation of 
            new object of this class.
        """
        keys_result = list(self._polynomial.keys()) + list(other._polynomial.keys())
        result = dict.fromkeys(keys_result)
        for key in result.keys():
            result[key] = self._polynomial.get(key, 0) - other._polynomial.get(key, 0)
        return DictPolynomial(result)

    def __mul__(self, other):
        """ Multiplication of two polynomials and creation of 
            new object of this class.
        """
        result = defaultdict(int)
        for k1, v1 in self._polynomial.items():
            for k2, v2 in other._polynomial.items():
                    result[k1 + k2] += v1*v2
        return DictPolynomial(result)

    def derivative(self):
        """ Calculating derivative of polynomial.
        """
        result = {}
        for k, v in self._polynomial.items():
            if k!=0:
                result[k-1] = v*k
        return DictPolynomial(result)

    def __divmod__(self, other):
        """ Division with the rest of two polynomials and creation of 
            new object of this class.
        """
        r = self
        result = DictPolynomial({})
        deg_p1 = max(r._polynomial.keys())
        deg_p2 = max(other._polynomial.keys())
        
        while deg_p1 >= deg_p2:
            if r._polynomial[deg_p1] % other._polynomial[deg_p2] == 0:
                new = DictPolynomial({deg_p1-deg_p2:int(r._polynomial[deg_p1]/other._polynomial[deg_p2])})
                result = result + new
                bottom = new * other
                r = r - bottom
                r._polynomial = {k:v for k,v in r._polynomial.items() if v!=0}
                if r._polynomial == {}:
                    break
                else:
                    deg_p1 = max(r._polynomial.keys())
            else:
                raise ValueError("Polynomials are indivisible")
        return result, r

    def __call__(self, number):
        """ Method overrides calling class which calculate value of polynomial
            in specific point using Horner's Method.
        """
        result = 0
        for i in range(max(self._polynomial.keys()), -1, -1):
            if i not in self._polynomial.keys():
                result = number*result
            else:
                result = number*result + self._polynomial[i] # schemat Hornera
        return result

    def __sub_roots(self, result, p, q, quotient_prev):
        """ Auxiliary function for calculating roots.
        """
        for comb in itertools.product(p, q):
            if comb[0]/comb[1] in result or Fraction(comb[0], comb[1]) in result:
                continue
            multi = DictPolynomial({0:-comb[0], 1:comb[1]})
            try:
                quotient, r = divmod(quotient_prev, multi)
            except:
                continue
            while not r._polynomial: # jeśli reszta z dzielenia ma same zera to znaczy, że badamy krotność pierwiastka
                try:
                    quotient, r = divmod(quotient, multi)
                except:
                    if comb[0] % comb[1] == 0:
                        result.append(int(comb[0]/comb[1]))
                    else:
                        result.append(Fraction(comb[0], comb[1]))
                    break
                if comb[0] % comb[1] == 0:
                    result.append(int(comb[0]/comb[1]))
                else:
                    result.append(Fraction(comb[0], comb[1]))     
        return result 

    def roots(self):
        """ Main method for calculating roots of polynomial.
            Returns list of rational roots.
        """
        result = []
        if 0 in self._polynomial.keys():
            p = self.findDivisors(self._polynomial[0])
            q = self.findDivisors(self._polynomial[max(self._polynomial.keys())])
            result = self.__sub_roots(result, p, q, self)                
        else: # w przypadku, gdy wyraz wolny jest 0
            multi = DictPolynomial("x")
            quotient, r = divmod(self, multi) # dzielenie przez (x-a)^n
            while not r._polynomial: # jeśli reszta z dzielenia jest pusta to znaczy, że badamy krotność pierwiastka
                quotient_prev = quotient
                quotient, r = divmod(quotient, multi) # dzielenie przez (x-a)^n 
                result.append(0)
            p = self.findDivisors(quotient_prev._polynomial[0])
            q = self.findDivisors(quotient_prev._polynomial[max(quotient_prev._polynomial.keys())])      
            result = self.__sub_roots(result, p, q, quotient_prev)
        return result

    def factorization(self):
        """ Main method for factorization of polynomials. Returns list
            of objects of this class, which are the factors of polynomial.
        """
        roots = self.roots() # wyznaczenie pierwiastków 
        l = len(roots)
        if l > 0:
            roots = [DictPolynomial({0:-i, 1:1}) if not isinstance(i, Fraction) 
            else DictPolynomial({0:-i.numerator, 1:i.denominator}) for i in roots] # utworzenie wielomianów z pierwiastków
            multi = DictPolynomial({0:1})
            if l > 1:
                for i in range(l):
                    multi = multi * roots[i] # wymnażanie kolejnych pierwiastków
                factor, _ = divmod(self, multi) # otrzymujemy wielomian poprzez dzielenie
            else:
                factor, _ = divmod(self, roots[0])
            if max(factor._polynomial.keys()) < 4:
                if factor._polynomial != {0:1}: # nie dodajemy 1 jako wyniku dzielenia
                    roots.append(factor) # do pierwiastków dodajemy nierozkłdalną część
            else:
                factors = factor.__no_roots_factorization()
                roots += factors
        else:
            if max(self._polynomial.keys()) < 4:
                roots = self # do pierwiastków dodajemy nierozkłdalną część
            else:
                factors = self.__no_roots_factorization()
                roots += factors
        return roots

    def __check_degree(self, list_of_polynomials):
        """ Checks whether the factorization still makes sense, 
            i.e. that the degree must be greater than 3
        """
        for poly in list_of_polynomials:
            if max(poly._polynomial.keys()) > 3:
                return True
        return False

    def __no_roots_factorization(self):
        """ Auxiliary function for calculating factorization.
            Implements Kronecker's Method.
        """
        factor_polynomials = [self]
        checked_polynomials = []
        while self.__check_degree(factor_polynomials):
            found = False
            for poly in factor_polynomials: # znajdz wielomian o rzędzie większym niż 3
                if max(poly._polynomial.keys()) > 3:
                    currentPoly = poly
                    break
            l = max(currentPoly._polynomial.keys()) + 1
            k = mt.floor(l/2) if l % 2 == 0 else mt.floor(l/2) + 1
            values = [] # kolejne wartosci wielomianu
            equations = [] # rownania w zaleznosci od wybranych punktow
            for i in range(k): # obliczanie macierzy do rownan oraz wartosci wielomianu
                values.append(currentPoly(i))
                eq = [i**p for p in range(k)]
                equations.append(eq)
            A = np.array(equations)
            inv_A = np.linalg.inv(A)
            factors_of_values = []
            for number in values: # wyznaczanie dzielnikow kolejnych wartosci wielomianu
                factors_of_values.append(self.findDivisors(number))
            without_minus = []
            for y in factors_of_values[0]: # usuwanie tylko tych kombinacji ktore roznia sie tylko znakiem
                if y > 0:
                    without_minus.append(y)
            
            factors_of_values[0] = without_minus
            
            for combination in itertools.product(*factors_of_values):
                combination = list(combination)
                
                B = np.array(combination) # macierz B potrzebna do znalezienia współczynników wielomianu
                X = inv_A.dot(B) 
                X = {i:int(np.round(X[i])) for i in range(X.size) if X[i] != 0}

                X = DictPolynomial(X)
                try:
                    factor, r = divmod(currentPoly, X) # sprawdzenie czy wielomian jest dzielnikiem
                except:
                    continue
                if r._polynomial=={} and max(X._polynomial.keys()) > 0: # jesli znalezlismy dzielnik to dodajemy go do rozkladu wielomianu wraz z dzielnikiem, który idzie z nim w parze
                    factor_polynomials.append(X)
                    factor_polynomials.append(factor)
                    found = True
                    break
            if found: 
                factor_polynomials.remove(currentPoly) # jeśli znaleźliśmy rozkład to usuwamy wielomian, którym się zajmowaliśmy
            else:
                factor_polynomials.remove(currentPoly) # jeśli brak rozkładu dla wielomianu to usuwamy go z wielomianów do sprawdzenia i dodajemy do już sprawdzonych
                checked_polynomials.append(currentPoly)
        factor_polynomials += checked_polynomials
        return factor_polynomials


from setuptools import setup, find_packages


setup(
    name='intpolylib',
    version='0.2',
    author="Dariusz Libecki",
    author_email='dariuszlibecki@gmail.com',
    license='MIT',
    long_description='README.md',
    long_description_content_type="text/markdown",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/Kerad20/polylib',
    keywords='polynomials',
    install_requires=[
          'numpy',
      ],

)